import numpy as np
import pandas as pd

import json
import time
import pytz
import inspect
import sys

from xone import __version__

__all__ = ['__version__']

DEFAULT_TZ = pytz.FixedOffset(-time.timezone / 60)


def tolist(iterable):
    """
    Simpler implementation of flatten method

    Args:
        iterable: any array or value

    Returns:
        list: list of unique values

    Examples:
        >>> tolist('xyz')
        ['xyz']
        >>> tolist(['ab', 'cd', 'xy', 'ab'])
        ['ab', 'cd', 'xy']
    """
    return pd.Series(iterable).drop_duplicates().tolist()


def fmt_dt(dt, fmt='%Y-%m-%d'):
    """
    Format date string

    Args:
        dt: any date format
        fmt: output date format

    Returns:
        str: date format

    Examples:
        >>> fmt_dt(dt='2018-12')
        '2018-12-01'
        >>> fmt_dt(dt='2018-12-31', fmt='%Y%m%d')
        '20181231'
    """
    return pd.Timestamp(dt).strftime(fmt)


def trade_day(dt, cal='US'):
    """
    Latest trading day w.r.t given dt

    Args:
        dt: date of reference
        cal: trading calendar

    Returns:
        pd.Timestamp: last trading day

    Examples:
        >>> trade_day('2018-12-25', cal='US').strftime('%Y-%m-%d')
        '2018-12-24'
    """
    from xone import calendar

    dt = pd.Timestamp(dt).date()
    return calendar.trading_dates(start=dt - pd.Timedelta('10D'), end=dt, cal=cal)[-1]


def cur_time(typ='date', tz=DEFAULT_TZ, trading=True, cal='US'):
    """
    Current time

    Args:
        typ: one of ['date', 'time', 'time_path', 'raw', '']
        tz: timezone
        trading: check if current date is trading day
        cal: trading calendar

    Returns:
        relevant current time or date

    Examples:
        >>> cur_dt = pd.Timestamp('now')
        >>> cur_time(typ='date', trading=False) == cur_dt.strftime('%Y-%m-%d')
        True
        >>> cur_time(typ='time', trading=False) == cur_dt.strftime('%Y-%m-%d %H:%M:%S')
        True
        >>> cur_time(typ='time_path', trading=False) == cur_dt.strftime('%Y-%m-%d/%H-%M-%S')
        True
        >>> isinstance(cur_time(typ='raw', tz='Europe/London'), pd.Timestamp)
        True
        >>> isinstance(cur_time(typ='raw', trading=True), pd.Timestamp)
        True
        >>> cur_time(typ='', trading=False) == cur_dt.date()
        True
    """
    dt = pd.Timestamp('now', tz=tz)

    if typ == 'date':
        if trading:
            return trade_day(dt=dt, cal=cal).strftime('%Y-%m-%d')
        return dt.strftime('%Y-%m-%d')

    if typ == 'time': return dt.strftime('%Y-%m-%d %H:%M:%S')
    if typ == 'time_path': return dt.strftime('%Y-%m-%d/%H-%M-%S')
    if typ == 'raw': return dt

    return trade_day(dt).date() if trading else dt.date()


def align_data(*args):
    """
    Resample and aligh data for defined frequency

    Args:
        *args: DataFrame of data to be aligned

    Returns:
        pd.DataFrame: aligned data with renamed columns

    Examples:
        >>> start = '2018-09-10T10:10:00'
        >>> tz = 'Australia/Sydney'
        >>> idx = pd.date_range(start=start, periods=6, freq='min').tz_localize(tz)
        >>> close_1 = [31.08, 31.10, 31.11, 31.07, 31.04, 31.04]
        >>> vol_1 = [10166, 69981, 14343, 10096, 11506, 9718]
        >>> d1 = pd.DataFrame(dict(price=close_1, volume=vol_1), index=idx)
        >>> d1
                                   price  volume
        2018-09-10 10:10:00+10:00  31.08   10166
        2018-09-10 10:11:00+10:00  31.10   69981
        2018-09-10 10:12:00+10:00  31.11   14343
        2018-09-10 10:13:00+10:00  31.07   10096
        2018-09-10 10:14:00+10:00  31.04   11506
        2018-09-10 10:15:00+10:00  31.04    9718
        >>> close_2 = [70.81, 70.78, 70.85, 70.79, 70.79, 70.79]
        >>> vol_2 = [4749, 6762, 4908, 2002, 9170, 9791]
        >>> d2 = pd.DataFrame(dict(price=close_2, volume=vol_2), index=idx)
        >>> d2
                                   price  volume
        2018-09-10 10:10:00+10:00  70.81    4749
        2018-09-10 10:11:00+10:00  70.78    6762
        2018-09-10 10:12:00+10:00  70.85    4908
        2018-09-10 10:13:00+10:00  70.79    2002
        2018-09-10 10:14:00+10:00  70.79    9170
        2018-09-10 10:15:00+10:00  70.79    9791
        >>> align_data(d1, d2)
                                   price_1  volume_1  price_2  volume_2
        2018-09-10 10:10:00+10:00    31.08     10166    70.81      4749
        2018-09-10 10:11:00+10:00    31.10     69981    70.78      6762
        2018-09-10 10:12:00+10:00    31.11     14343    70.85      4908
        2018-09-10 10:13:00+10:00    31.07     10096    70.79      2002
        2018-09-10 10:14:00+10:00    31.04     11506    70.79      9170
        2018-09-10 10:15:00+10:00    31.04      9718    70.79      9791
    """
    res = pd.DataFrame(pd.concat([
        d.loc[~d.index.duplicated(keep='first')].rename(
            columns=lambda vv: '%s_%d' % (vv, i + 1)
        ) for i, d in enumerate(args)
    ], axis=1))
    data_cols = [col for col in res.columns if col[-2:] == '_1']
    other_cols = [col for col in res.columns if col[-2:] != '_1']
    res.loc[:, other_cols] = res.loc[:, other_cols].fillna(method='pad')
    return res.dropna(subset=data_cols)


def cat_data(data_kw):
    """
    Concatenate data with ticker as sub column index

    Args:
        data_kw: key = ticker, value = pd.DataFrame

    Returns:
        pd.DataFrame

    Examples:
        >>> start = '2018-09-10T10:10:00'
        >>> tz = 'Australia/Sydney'
        >>> idx = pd.date_range(start=start, periods=6, freq='min').tz_localize(tz)
        >>> close_1 = [31.08, 31.10, 31.11, 31.07, 31.04, 31.04]
        >>> vol_1 = [10166, 69981, 14343, 10096, 11506, 9718]
        >>> d1 = pd.DataFrame(dict(price=close_1, volume=vol_1), index=idx)
        >>> close_2 = [70.81, 70.78, 70.85, 70.79, 70.79, 70.79]
        >>> vol_2 = [4749, 6762, 4908, 2002, 9170, 9791]
        >>> d2 = pd.DataFrame(dict(price=close_2, volume=vol_2), index=idx)
        >>> sample = cat_data({'BHP AU': d1, 'RIO AU': d2})
        >>> sample.columns
        MultiIndex([('BHP AU',  'price'),
                    ('BHP AU', 'volume'),
                    ('RIO AU',  'price'),
                    ('RIO AU', 'volume')],
                   names=['ticker', None])
        >>> r = sample.transpose().iloc[:, :2]
        >>> r.index.names = (None, None)
        >>> r
                       2018-09-10 10:10:00+10:00  2018-09-10 10:11:00+10:00
        BHP AU price                       31.08                      31.10
               volume                  10,166.00                  69,981.00
        RIO AU price                       70.81                      70.78
               volume                   4,749.00                   6,762.00
    """
    if len(data_kw) == 0: return pd.DataFrame()
    return pd.DataFrame(
        pd.concat([
            data
            .assign(ticker=ticker)
            .set_index('ticker', append=True)
            .unstack('ticker')
            .swaplevel(0, 1, axis=1)
            for ticker, data in data_kw.items()
        ], axis=1)
    )


def flatten(iterable, maps=None, unique=False):
    """
    Flatten any array of items to list

    Args:
        iterable: any array or value
        maps: map items to values
        unique: drop duplicates

    Returns:
        list: flattened list

    References:
        https://stackoverflow.com/a/40857703/1332656

    Examples:
        >>> flatten('abc')
        ['abc']
        >>> flatten(1)
        [1]
        >>> flatten(1.)
        [1.0]
        >>> flatten(['ab', 'cd', ['xy', 'zz']])
        ['ab', 'cd', 'xy', 'zz']
        >>> flatten(['ab', ['xy', 'zz']], maps={'xy': '0x'})
        ['ab', '0x', 'zz']
    """
    if iterable is None: return []
    if maps is None: maps = dict()

    if isinstance(iterable, (str, int, float)):
        return [maps.get(iterable, iterable)]

    else:
        x = [maps.get(item, item) for item in _to_gen_(iterable)]
        return list(set(x)) if unique else x


def _to_gen_(iterable):
    """
    Recursively iterate lists and tuples
    """
    from collections.abc import Iterable

    for elm in iterable:
        if isinstance(elm, Iterable) and not isinstance(elm, (str, bytes)):
            yield from flatten(elm)
        else: yield elm


def to_frame(data_list, exc_cols=None, **kwargs):
    """
    Dict in Python 3.6 keeps insertion order, but cannot be relied upon
    This method is to keep column names in order
    In Python 3.7 this method is redundant

    Args:
        data_list: list of dict
        exc_cols: exclude columns

    Returns:
        pd.DataFrame

    Example:
        >>> d_list = [
        ...     dict(sid=1, symbol='1 HK', price=89),
        ...     dict(sid=700, symbol='700 HK', price=350)
        ... ]
        >>> to_frame(d_list)
           sid  symbol  price
        0    1    1 HK     89
        1  700  700 HK    350
        >>> to_frame(d_list, exc_cols=['price'])
           sid  symbol
        0    1    1 HK
        1  700  700 HK
    """
    from collections import OrderedDict

    # noinspection PyTypeChecker
    return pd.DataFrame(
        pd.Series(data_list)
        .apply(OrderedDict)
        .tolist(),
        **kwargs
    ).drop(columns=[] if exc_cols is None else exc_cols)


def read_zip(zip_url: str, read_func, **kwargs) -> pd.DataFrame:
    """
    Read zip file, either from url or local

    Args:
        zip_url: URL or local address
        read_func: function
        **kwargs: key-word arguments passed to read_func

    Returns:
        pd.DataFrame

    Examples:
        >>> from xone.files import abspath
        >>> test_folder = f'{abspath(__file__)}/tests/files'
        >>> sample = read_zip(
        ...     zip_url=f'{test_folder}/master.zip',
        ...     read_func=pd.read_csv,
        ...     skiprows=list(range(9)) + [10],
        ...     sep='|',
        ...     encoding='latin1',
        ... )
        >>> (
        ...     sample[sample['Company Name'].str.startswith('BP')]
        ...     .iloc[:, :-1]
        ...     .reset_index(drop=True)
        ... )
               CIK                    Company Name Form Type  Date Filed
        0  1167581          BP CAPITAL MARKETS PLC    25-NSE  2020-10-02
        1  1167583  BP CAPITAL MARKETS AMERICA INC    25-NSE  2020-10-02
        2  1798039              BPGIC HOLDINGS Ltd  SC 13D/A  2020-10-05
        3   313807                          BP PLC       6-K  2020-10-05
        4   850033    BP PRUDHOE BAY ROYALTY TRUST       8-K  2020-10-02
    """
    import tempfile
    import zipfile
    import urllib.parse
    import urllib.request

    data_list = []
    info = urllib.parse.urlparse(zip_url)
    is_url = len(info.scheme) > 1
    with tempfile.TemporaryFile(mode='w+b') as tmp:
        if is_url: tmp.write(urllib.request.urlopen(zip_url).read())
        with zipfile.ZipFile(tmp if is_url else zip_url) as zf:
            for f_name in zf.namelist():
                with zf.open(f_name) as z:
                    try:
                        data_list.append(read_func(z, **kwargs))
                    except Exception as e:
                        print(e)

    return pd.DataFrame(pd.concat(data_list, sort=False))


def func_scope(func):
    """
    Function scope name

    Args:
        func: python function

    Returns:
        str: module_name.func_name

    Examples:
        >>> func_scope(flatten)
        'xone.utils.flatten'
        >>> func_scope(json.dump)
        'json.dump'
    """
    cur_mod = sys.modules[func.__module__]
    return f'{cur_mod.__name__}.{func.__name__}'


def func_kwarg(func, **kwargs) -> dict:
    """
    Pass kwargs to function only if they are required

    Args:
        func: function

    Returns:
        dict

    Examples:
        >>> func_kwarg(func=read_zip, red=1)
        {'red': 1}
        >>> func_kwarg(func=func_scope, red=1)
        {}
    """
    if not callable(func): return {}

    param = inspect.signature(func).parameters
    kind = pd.Series({k: v.kind for k, v in param.items()})
    if kind.max() == 4: return kwargs
    return {k: v for k, v in kwargs.items() if k in kind}


def perf(data: (pd.DataFrame, pd.Series)) -> (pd.DataFrame, pd.Series):
    """
    Price performance based at 100

    Args:
        data: price series or dataframe

    Returns:
        pd.DataFrame or pd.Series

    Examples:
        >>> (
        ...     pd.DataFrame({
        ...         'P1': [1, 1.03, 1.02, 1.06],
        ...         'P2': [1, .99, .97, .97],
        ...     })
        ...     .pipe(perf)
        ... )
              P1     P2
        0 100.00 100.00
        1 103.00  99.00
        2 102.00  97.00
        3 106.00  97.00
    """
    if isinstance(data, pd.Series):
        return (
            data
            .dropna()
            .pct_change()
            .fillna(0)
            .add(1)
            .cumprod()
            .mul(100)
        )
    return data.apply(perf, axis=0)


def format_float(digit=0, is_pct=False):
    """
    Number display format for pandas

    Args:
        digit: number of digits to keep
               if negative, add one space in front of positive pct
        is_pct: % display

    Returns:
        lambda function to format floats

    Examples:
        >>> format_float(0)(1e5)
        '100,000'
        >>> format_float(1)(1e5)
        '100,000.0'
        >>> format_float(-1, True)(.2)
        ' 20.0%'
        >>> format_float(-1, True)(-.2)
        '-20.0%'
        >>> pd.options.display.float_format = format_float(2)
    """
    if is_pct:
        space = ' ' if digit < 0 else ''
        fmt = f'{{:{space}.{abs(int(digit))}%}}'
        return lambda vv: 'NaN' if np.isnan(vv) else fmt.format(vv)

    return lambda vv: 'NaN' if np.isnan(vv) else (
        f'{{:,.{digit}f}}'.format(vv) if vv else '-' + ' ' * abs(digit)
    )


class FString(object):

    def __init__(self, str_fmt):
        self.str_fmt = str_fmt

    def __str__(self):
        kwargs = inspect.currentframe().f_back.f_globals.copy()
        kwargs.update(inspect.currentframe().f_back.f_locals)
        return self.str_fmt.format(**kwargs)


def fstr(fmt, **kwargs):
    """
    Delayed evaluation of f-strings

    Args:
        fmt: f-string but in terms of normal string, i.e., '{path}/{file}.parq'
        **kwargs: variables for f-strings, i.e., path, file = '/data', 'daily'

    Returns:
        FString object

    References:
        https://stackoverflow.com/a/42497694/1332656
        https://stackoverflow.com/a/4014070/1332656

    Examples:
        >>> use_fmt = '{data_path}/{data_file}.parq'
        >>> fstr(fmt=use_fmt, data_path='your/data/path', data_file='sample')
        'your/data/path/sample.parq'
    """
    locals().update(kwargs)
    return f'{FString(str_fmt=fmt)}'


def to_str(data: dict, fmt='{key}={value}', sep=', ', public_only=True):
    """
    Convert dict to string

    Args:
        data: dict
        fmt: how key and value being represented
        sep: how pairs of key and value are seperated
        public_only: if display public members only

    Returns:
        str: string representation of dict

    Examples:
        >>> test_dict = dict(b=1, a=0, c=2, _d=3)
        >>> to_str(test_dict)
        '{b=1, a=0, c=2}'
        >>> to_str(test_dict, sep='|')
        '{b=1|a=0|c=2}'
        >>> to_str(test_dict, public_only=False)
        '{b=1, a=0, c=2, _d=3}'
    """
    if public_only: keys = list(filter(lambda vv: vv[0] != '_', data.keys()))
    else: keys = list(data.keys())
    return '{' + sep.join([
        to_str(data=v, fmt=fmt, sep=sep)
        if isinstance(v, dict) else fstr(fmt=fmt, key=k, value=v)
        for k, v in data.items() if k in keys
    ]) + '}'


def inst_repr(instance, fmt='str', public_only=True):
    """
    Generate class instance signature from its __dict__
    From python 3.6 dict is ordered and order of attributes will be preserved automatically

    Args:
        instance: class instance
        fmt: ['json', 'str']
        public_only: if display public members only

    Returns:
        str: string or json representation of instance

    Examples:
        >>> inst_repr(1)
        ''
        >>> class SampleClass(object):
        ...     def __init__(self):
        ...         self.b = 3
        ...         self.a = 4
        ...         self._private_ = 'hidden'
        >>>
        >>> s = SampleClass()
        >>> inst_repr(s)
        '{b=3, a=4}'
        >>> inst_repr(s, public_only=False)
        '{b=3, a=4, _private_=hidden}'
        >>> json.loads(inst_repr(s, fmt='json'))
        {'b': 3, 'a': 4}
        >>> inst_repr(s, fmt='unknown')
        ''
    """
    if not hasattr(instance, '__dict__'): return ''

    if public_only:
        inst_dict = {k: v for k, v in instance.__dict__.items() if k[0] != '_'}
    else: inst_dict = instance.__dict__

    if fmt == 'json': return json.dumps(inst_dict, indent=2)
    if fmt == 'str': return to_str(inst_dict, public_only=public_only)

    return ''


def load_module(full_path):
    """
    Load module from full path

    Args:
        full_path: module full path name

    Returns:
        python module

    References:
        https://stackoverflow.com/a/67692/1332656

    Examples:
        >>> import os
        >>>
        >>> cur_file = os.path.abspath(__file__).replace('\\\\', '/')
        >>> cur_path = '/'.join(cur_file.split('/')[:-1])
        >>> load_module(f'{cur_path}/files.py').__name__
        'files'
        >>> load_module(f'{cur_path}/files.pyc')
        Traceback (most recent call last):
        ImportError: not a python file: files.pyc
    """
    from importlib import util

    file_name = full_path.replace('\\', '/').split('/')[-1]
    if file_name[-3:] != '.py':
        raise ImportError(f'not a python file: {file_name}')
    module_name = file_name[:-3]

    spec = util.spec_from_file_location(name=module_name, location=full_path)
    module = util.module_from_spec(spec=spec)
    spec.loader.exec_module(module=module)

    return module


class AttributeDict(dict):
    """
    Dot access support for dict attributes

    References:
        https://stackoverflow.com/a/5021467/1332656
    """
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__



import requests
import time

from xone import logs
from collections import namedtuple
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.common.exceptions import \
    NoSuchElementException, ElementClickInterceptedException


BROWSERS = {
    'firefox': (webdriver.Firefox, webdriver.FirefoxOptions()),
    'chrome': (webdriver.Chrome, webdriver.ChromeOptions()),
}

PageSource = namedtuple('PageSource', ['source', 'soup'])


def get_text(elem) -> str:
    """
    Get text from element
    """
    for line in elem.find_all('br'): line.replace_with('\n')
    return elem.text.strip()


def to_text(soup, elem_name, func: callable = None) -> list:
    """
    Get list of texts from row data
    """
    return [
        func(get_text(elem=elem)) if callable(func) else get_text(elem=elem)
        for elem in soup.select(elem_name)
    ]


def get_browser(browser, headless=True, *args):
    """
    Get browser
    """
    br, br_opt = BROWSERS[browser]
    if headless: br_opt.add_argument('--headless')
    br_opt.add_argument('--disable-gpu')
    for arg in args: br_opt.add_argument(arg)
    return br(options=br_opt)


def page_source(
        url: str,
        browser: str = '',
        click: str = '',
        headless=True,
        *args
) -> PageSource:
    """
    Page source from url

    Args:
        url: url
        browser: one of firefox, chrome and ie
        click: element to click - useful for scroll down page
               will use browser if given - default browser Chrome
        headless: whether to show browser

    Returns:
        PageSource
    """
    logger = logs.get_logger(page_source)

    if click and (not browser): browser = 'chrome'
    if not browser:
        page = requests.get(url)
        if page.status_code == 200:
            return PageSource(
                source=page.content,
                soup=BeautifulSoup(page.content, 'lxml'),
            )
        else:
            logger.warning(f'Invalid status code {page.status_code} for url:\n{url}')
            return PageSource(source='', soup=BeautifulSoup('', 'lxml'))

    if browser not in BROWSERS:
        raise LookupError(
            f'Cannot find browser: {browser}.\n'
            f'Valid browsers: {list(BROWSERS.keys())}.'
        )

    with get_browser(browser, headless=headless, *args) as driver:
        driver.get(url=url)
        if click:
            try:
                driver.find_element_by_xpath(click).click()
            except (NoSuchElementException, ElementClickInterceptedException) as e:
                logger.error(e)
            time.sleep(1)
        return PageSource(
            source=driver.page_source,
            soup=BeautifulSoup(driver.page_source, 'lxml'),
        )



import pandas as pd

from xone import utils
from matplotlib import pyplot as plt


def plot_ts(
        data: (pd.Series, pd.DataFrame), fld='close',
        tz=None, vline=None, **kwargs
):
    """
    Time series data plots

    Args:
        data: data in pd.Series or pd.DataFrame
        fld: which field to plot for multi-index
        tz: tz info - applied to merged time series data
        vline: kwargs for verticle lines

    Returns:
        matplotlib plot
    """
    to_plot = data
    if isinstance(data, pd.DataFrame):
        if isinstance(data.columns, pd.MultiIndex) and \
                (fld in data.columns.get_level_values(1)):
            to_plot = data.xs(fld, axis=1, level=1)
        elif fld in data.columns:
            to_plot = data[fld]

    if isinstance(to_plot, pd.Series):
        pl_val = pd.Series(data.values, name=data.name)
    else:
        pl_val = pd.DataFrame(data.values, columns=data.columns)

    # Proper raw datetime index
    idx = data.index
    if isinstance(idx, pd.MultiIndex):
        for n in range(len(idx.levels))[::-1]:
            sidx = idx.get_level_values(n)
            if isinstance(sidx, pd.DatetimeIndex):
                idx = sidx
                break

    # Standardize timezone
    assert isinstance(idx, pd.DatetimeIndex), idx
    if tz is not None: idx = idx.tz_convert(tz)

    raw_idx = idx.to_series(keep_tz=True).reset_index(drop=True)
    dt_idx = pd.Series(raw_idx.dt.date.unique())
    diff = raw_idx.diff()

    is_day = diff.min() >= pd.Timedelta(days=1)
    num_days = dt_idx.size

    if num_days >= kwargs.pop('month_min_cnt', 90):
        # Monthly ticks
        xticks = raw_idx.loc[raw_idx.dt.month.diff().ne(0)]
    elif num_days >= kwargs.pop('week_min_cnt', 15):
        # Weekly ticks
        xticks = raw_idx.loc[raw_idx.dt.weekofyear.diff().ne(0)]
    elif is_day:
        xticks = raw_idx.index
    else:
        # Daily ticks - to be improved
        xticks = raw_idx.loc[raw_idx.dt.day.diff().ne(0)]

    # Plot
    ax = pl_val.plot(**kwargs)
    plt.xticks(ticks=xticks.index.tolist(), labels=xticks.dt.date, rotation=30)

    if not isinstance(vline, dict): vline = dict()
    vline['color'] = vline.get('color', '#FFFFFF')
    vline['linestyle'] = vline.get('linestyle', '-')
    vline['linewidth'] = vline.get('linewidth', 1)
    for xc in xticks: plt.axvline(x=xc, **vline)

    return ax


def plot_multi(data, cols=None, spacing=.06, color_map=None, plot_kw=None, **kwargs):
    """
    Plot data with multiple scaels together

    Args:
        data: DataFrame of data
        cols: columns to be plotted
        spacing: spacing between legends
        color_map: customized colors in map
        plot_kw: kwargs for each plot
        **kwargs: kwargs for the first plot

    Returns:
        ax for plot
    """
    from pandas import plotting

    if cols is None: cols = data.columns
    if plot_kw is None: plot_kw = [{}] * len(cols)
    if len(cols) == 0: return
    num_colors = len(utils.flatten(cols))

    # Get default color style from pandas
    if hasattr(plotting, '_matplotlib'):
        style = getattr(plotting, '_matplotlib').style
    else:
        style = getattr(plotting, '_style')
    colors = getattr(style, '_get_standard_colors')(num_colors=num_colors)
    if color_map is None: color_map = dict()

    fig = plt.figure()
    ax, lines, labels, c_idx = None, [], [], 0
    for n, col in enumerate(cols):
        if isinstance(col, (list, tuple)):
            ylabel = ' / '.join(cols[n])
            color = [
                color_map.get(cols[n][_ - c_idx], colors[_ % len(colors)])
                for _ in range(c_idx, c_idx + len(cols[n]))
            ]
            c_idx += len(col)
        else:
            ylabel = col
            color = color_map.get(col, colors[c_idx % len(colors)])
            c_idx += 1
        if 'color' in plot_kw[n]: color = plot_kw[n].pop('color')

        if ax is None:
            # First y-axes
            legend = plot_kw[0].pop('legend', kwargs.pop('legend', False))
            ax = data.loc[:, col].plot(
                label=col, color=color, legend=legend, zorder=n, **plot_kw[0], **kwargs
            )
            ax.set_ylabel(ylabel=ylabel)
            line, label = ax.get_legend_handles_labels()
            ax.spines['left'].set_edgecolor('#D5C4A1')
            ax.spines['left'].set_alpha(.5)

        else:
            # Multiple y-axes
            legend = plot_kw[n].pop('legend', False)
            ax_new = ax.twinx()
            ax_new.spines['right'].set_position(('axes', 1 + spacing * (n - 1)))
            data.loc[:, col].plot(
                ax=ax_new, label=col, color=color, legend=legend, zorder=n, **plot_kw[n]
            )
            ax_new.set_ylabel(ylabel=ylabel)
            line, label = ax_new.get_legend_handles_labels()
            ax_new.spines['right'].set_edgecolor('#D5C4A1')
            ax_new.spines['right'].set_alpha(.5)
            ax_new.grid(False)

        # Proper legend position
        lines += line
        labels += label

    fig.legend(lines, labels, loc=8, prop=dict(), ncol=num_colors).set_zorder(len(cols))
    ax.set_xlabel(' \n ')

    return ax


def plot_h(data, cols, wspace=.1, plot_kw=None, **kwargs):
    """
    Plot horizontally

    Args:
        data: DataFrame of data
        cols: columns to be plotted
        wspace: spacing between plots
        plot_kw: kwargs for each plot
        **kwargs: kwargs for the whole plot

    Returns:
        axes for plots
    """
    if plot_kw is None: plot_kw = [dict()] * len(cols)

    _, axes = plt.subplots(nrows=1, ncols=len(cols), **kwargs)
    plt.subplots_adjust(wspace=wspace)
    for n, col in enumerate(cols):
        data.loc[:, col].plot(ax=axes[n], **plot_kw[n])

    return axes


def add_lines(
        axes: plt.Axes, xs=None, ys=None, colors=None, **kwargs
) -> plt.Axes:
    """
    Add horizontal or vertical lines to charts

    Args:
        axes: axes to add
        xs: Xs
        ys: Ys
        colors: list of colors (Xs first then Ys if both given
        **kwargs: kwargs to pass

    Returns:
        plt.Axes
    """
    idx = 0
    if colors is None: colors = ['darkgreen', 'darkorange', 'darkred']

    if xs is not None:
        if isinstance(xs, str): xs = [xs]
        if not hasattr(xs, '__iter__'): xs = [xs]
        ylim = axes.get_ylim()
        for x in xs:
            axes.vlines(
                x=x, ymin=ylim[0], ymax=ylim[1],
                colors=colors[idx % len(colors)], **kwargs
            )
            idx = idx + 1
        axes.set_ylim(ymin=ylim[0], ymax=ylim[1])

    if ys is not None:
        if isinstance(ys, str): ys = [ys]
        if not hasattr(ys, '__iter__'): ys = [ys]
        xlim = axes.get_xlim()
        for y in ys:
            axes.hlines(
                y=y, xmin=xlim[0], xmax=xlim[1],
                colors=colors[idx % len(colors)], **kwargs
            )
            idx = idx + 1
        axes.set_xlim(xmin=xlim[0], xmax=xlim[1])

    return axes


def xticks(data: (pd.Series, pd.DataFrame), max_cnt=8) -> pd.DataFrame:
    """
    Get x-ticks for intraday data

    Args:
        data: intraday time series data
        max_cnt: max number of intervals

    Returns:
        pd.DataFrame
    """
    to_plot = data.sort_index()
    idx = (
        to_plot.index
        .to_series()
        .diff()
        .shift(-1)
        .fillna(pd.Timedelta('10 days'))
    )
    qtile = idx.quantile(q=.9)
    dates = idx[idx > qtile].index
    steps = int(dates.size / max_cnt) + 1
    res = dates[::-1][::steps][::-1]
    if dates[0] not in res: res = res.insert(0, dates[0])

    return pd.DataFrame(
        data={
            'ticks': res.to_series().apply(data.index.get_loc),
            'labels': res.strftime('%Y-%m-%d'),
        },
        index=res,
    )


def intraday(data: (pd.Series, pd.DataFrame), max_cnt=8, **kwargs):
    """
    Plot intraday data

    Args:
        data: intraday time series data
        max_cnt: max number of intervals
    """
    ax = data.reset_index(drop=True).plot(**kwargs)
    xrng = int(data.shape[0] * .01)
    ax.set_xlim(xmin=-xrng, xmax=data.shape[0] + xrng)
    ax.set_ylim(
        ymin=data.values.min() * .97,
        ymax=data.values.max() * 1.03,
    )
    ax.grid(which='major', axis='both', linestyle='--')
    plt.xticks(
        rotation=30,
        ha='right',
        rotation_mode='anchor',
        **xticks(data=data, max_cnt=max_cnt).to_dict('list'),
    )
    return ax



import pandas as pd

import sqlite3
import json

WAL_MODE = 'PRAGMA journal_mode=WAL'
ALL_TABLES = 'SELECT name FROM sqlite_master WHERE type="table"'


class Singleton(type):

    _instances_ = {}

    def __call__(cls, *args, **kwargs):
        # Default values for class init
        default_keys = ['db_file', 'keep_live']
        kw = {**dict(zip(default_keys, args)), **kwargs}
        kw['keep_live'] = kw.get('keep_live', False)

        # Singleton instance
        key = json.dumps(kw)
        if key not in cls._instances_:
            cls._instances_[key] = super(Singleton, cls).__call__(**kw)
        return cls._instances_[key]


class SQLite(metaclass=Singleton):
    """
    Examples:
        >>> from xone import files
        >>>
        >>> db_file_ = f'{files.abspath(__file__)}/tests/xone.db'
        >>> with SQLite(db_file_) as con_:
        ...     _ = con_.execute('DROP TABLE IF EXISTS xone')
        ...     _ = con_.execute('CREATE TABLE xone (rowid int)')
        >>> db_ = SQLite(db_file_)
        >>> db_.tables()
        ['xone']
        >>> db_.replace_into(table='xone', rowid=1)
        >>> db_.select(table='xone')
           rowid
        0      1
    """

    def __init__(self, db_file, keep_live=False):

        self.db_file = db_file
        self.keep_live = keep_live
        self._con_ = None

    def tables(self) -> list:
        """
        All tables within database
        """
        keep_live = self.is_live
        res = self.con.execute(ALL_TABLES).fetchall()
        if not keep_live: self.close()
        return [r[0] for r in res]

    def select(self, table: str, cond='', **kwargs) -> pd.DataFrame:
        """
        SELECT query
        """
        keep_live = self.is_live
        q_str = select(table=table, cond=cond, **kwargs)
        data = self.con.execute(q_str).fetchall()
        if not keep_live: self.close()
        return pd.DataFrame(data, columns=self.columns(table=table))

    def select_recent(
            self,
            table: str,
            dateperiod: str,
            date_col: str = 'modified_date',
            cond='',
            **kwargs
    ) -> pd.DataFrame:
        """
        Select recent

        Args:
            table: table name
            dateperiod: time period, e.g., 1M, 1Q, etc.
            date_col: column for time period
            cond: conditions
            **kwargs: other select criteria

        Returns:
            pd.DataFrame
        """
        cols = self.columns(table=table)
        if date_col not in cols: return pd.DataFrame()
        start_dt = (
            pd.date_range(
                end='today', freq=dateperiod, periods=2, normalize=True,
            )[0]
            .strftime('%Y-%m-%d')
        )
        return (
            self.select(table=table, cond=cond, **kwargs)
            .query(f'{date_col} >= {start_dt}')
            .reset_index(drop=True)
        )

    def columns(self, table: str):
        """
        Table columns
        """
        return [
            info[1] for info in (
                self.con.execute(f'PRAGMA table_info({table})').fetchall()
            )
        ]

    def replace_into(self, table: str, data: pd.DataFrame = None, **kwargs):
        """
        Replace records into table

        Args:
            table: table name
            data: DataFrame - if given, **kwargs will be ignored
            **kwargs: record values
        """
        if isinstance(data, pd.DataFrame):
            keep_live = self.is_live
            # noinspection PyTypeChecker
            data.apply(
                lambda row: self.replace_into(table=table, _live_=True, **row),
                axis=1
            )
        else:
            keep_live = self.is_live or kwargs.get('_live_', False)
            self.con.execute(replace_into(table=table, **{
                k: v for k, v in kwargs.items() if k != '_live_'
            }))
        if not keep_live: self.close()

    @property
    def is_live(self) -> bool:
        if not isinstance(self._con_, sqlite3.Connection):
            return False
        try:
            self._con_.execute(ALL_TABLES)
            return True
        except sqlite3.Error:
            return False

    @property
    def con(self) -> sqlite3.Connection:
        if not self.is_live:
            self._con_ = sqlite3.connect(self.db_file)
            self._con_.execute(WAL_MODE)
        return self._con_

    def close(self, keep_live=False):
        try:
            self._con_.commit()
            if not keep_live: self._con_.close()
        except sqlite3.ProgrammingError:
            pass
        except sqlite3.Error as e:
            print(e)

    def __enter__(self):
        return self.con.cursor()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(keep_live=self.keep_live)


def db_value(val) -> str:
    """
    Database value as in query string
    """
    if isinstance(val, str):
        return json.dumps(val.replace('\"', '').strip())
    return json.dumps(val)


def select(table: str, cond='', **kwargs) -> str:
    """
    Query string of SELECT statement

    Args:
        table: table name
        cond: conditions
        **kwargs: data as kwargs

    Examples:
        >>> q1 = select('daily', ticker='ES1 Index', price=3000)
        >>> q1.splitlines()[-2].strip()
        'ticker="ES1 Index" AND price=3000'
        >>> q2 = select('daily', cond='price > 3000', ticker='ES1 Index')
        >>> q2.splitlines()[-2].strip()
        'price > 3000 AND ticker="ES1 Index"'
        >>> q3 = select('daily', cond='price > 3000')
        >>> q3.splitlines()[-2].strip()
        'price > 3000'
        >>> select('daily')
        'SELECT * FROM daily'
    """
    all_cond = [cond] + [
        f'{key}={db_value(value)}'
        for key, value in kwargs.items()
    ]
    where = ' AND '.join(filter(bool, all_cond))
    s = f'SELECT * FROM {table}'
    if where:
        return f"""
            {s}
            WHERE
            {where}
        """
    else:
        return s


def replace_into(table: str, **kwargs) -> str:
    """
    Query string of REPLACE INTO statement

    Args:
        table: table name
        **kwargs: data as kwargs

    Examples:
        >>> query = replace_into('daily', ticker='ES1 Index', price=3000)
        >>> query.splitlines()[1].strip()
        'REPLACE INTO daily (ticker, price)'
        >>> query.splitlines()[2].strip()
        'VALUES ("ES1 Index", 3000)'
    """
    return f"""
        REPLACE INTO {table} ({', '.join(list(kwargs.keys()))})
        VALUES ({', '.join(map(db_value, list(kwargs.values())))})
    """
